import re


def contain_any_url(s): 
    return bool(re.search(r'://',s))


def contain_any_javadoc_tag(s): 
    return bool(re.search(r'@[a-zA-Z0-9]+',s))
    

def contain_any_non_English(s): 
    return bool(re.search(r'[^\x00-\xff]',s))


def not_contain_any_letter(s): 
    return not bool(re.search(r'[a-zA-Z]', s))


def less_than_three_words(s): 
    words = s.split()
    if (len(words)<=2):
        return True
    else:
        return False
    

def end_with_question_mark(s):
    if len(s) < 1:
        return False
    return s[-1] == '?'

def contain_html_tag(s): 
    return bool(re.search(r'</?[^>]+>', s))
    
def detach_brackets(s):
    return re.sub(r'\([^\)]*\)','',s)

def detach_html_tag(s):
    return re.sub(r'</?[^>]+>','',s)


rule2fuc = {
    'contain_any_url':contain_any_url,
    'contain_any_javadoc_tag':contain_any_javadoc_tag,
    'contain_any_non_English':contain_any_non_English,
    'not_contain_any_letter':not_contain_any_letter,
    'less_than_three_words':less_than_three_words,
    'end_with_question_mark':end_with_question_mark,
    'contain_html_tag':contain_html_tag,
    'detach_brackets':detach_brackets,
    'detach_html_tag':detach_html_tag,
}


def rule_filter(comments,selected_rules=[],defined_rule_dict={}):
    if not isinstance(comments, list):
        raise TypeError('comments must be a list')

    new_comments = []
    indexs = []
    rule2fuc.update(defined_rule_dict)

    if len(selected_rules) < 1:
        selected_rules = list(rule2fuc.keys())
    rule_set = selected_rules + list(defined_rule_dict.keys())

    for i,c in enumerate(comments):
        flag = False
        for rule in rule_set:
            if rule.startswith('detach'):
                result = rule2fuc[rule](c)
                if result is True:
                    continue
                elif result is False:
                    flag = True
                    break
                elif isinstance(result,str):
                    c = result
                else:
                    raise TypeError('Function must return True, False or String')

        for rule in rule_set:
            if not rule.startswith('detach') and rule2fuc[rule](c):
                flag = True
                break

        if not flag:
            new_comments.append(c)
            indexs.append(i)

    return new_comments,indexs
        



import torch
import random
import torch.nn as nn
from torch.autograd import Variable
import torch.nn.functional as F

def embed ():
    # 预训练的 GloVe 词向量文件
    glove_file = './glove_DATA/python_glove.txt'
    weights = []
    with open(glove_file, 'r', encoding='utf-8') as f:
        for line in f:
            line = line.strip().split(' ')
            word = line[0]
            vector = list(map(float, line[1:]))
            weights.append(vector)
    weights = torch.tensor(weights)
    return weights


# Encoder
# ------------------------------------------------------------------------------

# Encode into Z with mu and log_var

class Encoder(nn.Module):
    def __init__(self, input_size, emb_size, output_size, dropout_p=0.1, cuda=True):
        super(Encoder, self).__init__()
        self.cuda = cuda
        self.emb_size = emb_size
        self.embed = nn.Embedding(input_size, emb_size)
        # 使用预训练词向量
        # self.embed = nn.Embedding.from_pretrained(embed())
        self.self_attention = nn.MultiheadAttention(emb_size, num_heads=1, dropout=dropout_p)
        self.o2p = nn.Linear(emb_size, output_size * 2)

    def forward(self, input):
        embedded = self.embed(input)
        output, hidden = self.self_attention(embedded, embedded, embedded)
        output = output.mean(dim=1)
        ps = self.o2p(output)
        mu, logvar = torch.chunk(ps, 2, dim=1)
        z = self.sample(mu, logvar)
        return mu, logvar, z

    def sample(self, mu, logvar):
        eps = Variable(torch.randn(mu.size()))
        if self.cuda:
            eps = eps.cuda()
        std = torch.exp(logvar / 2.0)
        return mu + eps * std


# Decoder
# ------------------------------------------------------------------------------

# Decode from Z into sequence


class Decoder(nn.Module):
    def __init__(self, input_size, emb_size, output_size, n_layers=1, dropout_p=0.1, cuda=True):
        super(Decoder, self).__init__()
        self.SOS = 1
        self.cuda = cuda
        self.embed = nn.Embedding(output_size, input_size)
        # 使用预训练词向量
        # self.embed = nn.Embedding.from_pretrained(embed())
        self.output_size = output_size
        self.n_layers = n_layers

        self.self_attention = nn.MultiheadAttention(input_size, num_heads=1, dropout=dropout_p)

        self.output_projection = nn.Sequential(
            nn.Linear(input_size, emb_size),
            nn.Tanh(),
            nn.Linear(emb_size, input_size)
        )
        self.generator = nn.Linear(input_size, output_size)

    def forward(self, hidden, trg):
        max_len = trg.size(1)
        hidden = hidden.unsqueeze(0).repeat(self.n_layers, 1, 1)
        logit = []
        input = Variable(torch.LongTensor([self.SOS] * trg.size(0)))
        if self.cuda:
            input = input.cuda()
        for i in range(max_len):
            hidden, input = self.step(hidden, input)
            logit.append(input)
            use_teacher_forcing = random.random() <= 0.9 if self.training and i > 0 else 1
            if use_teacher_forcing:
                input = trg[:, i]
            else:
                input = input.argmax(dim=1)
        logit = torch.stack(logit, dim=1)
        return logit

    # def sample(self, hidden, trg):
    #     max_len = trg.size(1)
    #     hidden = hidden.unsqueeze(0).repeat(self.n_layers,1,1)
    #     logit = []
    #     for i in range(max_len):
    #         hidden, token_logit = self.step(hidden, input)
    #         input = token_logit.argmax(dim=1)
    #         logit.append(token_logit)
    #     logit = torch.stack(logit, dim=1)
    #     return logit

    def step(self, hidden, token):
        token_embedding = self.embed(token.unsqueeze(0))
        hidden, attn_output_weights = self.self_attention(token_embedding, hidden, hidden)
        output = self.output_projection(hidden.squeeze(0))
        token_logit = self.generator(output)
        return hidden, token_logit


# Container
# ------------------------------------------------------------------------------


class VAE(nn.Module):
    def __init__(self, vocab_size, emb_size, output_size, dropout=0, n_layers=1, cuda=True):
        super(VAE, self).__init__()
        self.encoder = Encoder(vocab_size, emb_size, output_size, dropout_p=dropout, cuda=cuda)
        self.decoder = Decoder(output_size, emb_size, vocab_size, dropout_p=dropout, n_layers=n_layers, cuda=cuda)

    def forward(self, inputs):
        m, l, z = self.encoder(inputs)
        decoded = self.decoder(z, inputs)
        return m, l, z, decoded


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cocorf",
    version="0.1",
    author_email="Zxqinn@gmail.com",
    description="Comment-based codebase refining framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=['torch==1.3.1','numpy','sklearn','nltk'], 
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)


import json
import torch
import tqdm
import numpy as np
import torch.nn as nn
import sys
import os
sys.path.append(os.path.dirname(__file__))
from torch.utils.data import Dataset
from sklearn.mixture import GaussianMixture
import nltk
nltk.download('wordnet')
nltk.download('omw-1.4')
from nltk.stem import WordNetLemmatizer 
from multihead_model import *



class QueryDataset(Dataset):
    PAD = 0
    UNK = 1
    BOS = 2
    EOS = 3

    def __init__(self, query, word_vocab, query_len, is_docstring=False):
        super(QueryDataset, self).__init__()
        self.query_len = query_len
        self.word_vocab = word_vocab
        self.query = []
        lemmatizer = WordNetLemmatizer()  
        
        if isinstance(query,str):
            with open(query, 'r') as f:
                lines = f.readlines()
        else:
            lines = query
            
        for q in lines:
            words = q.lower().split()[:query_len]
            if is_docstring:
                if len(words)>0 and word_vocab.get(words[0],self.UNK) == self.UNK:
                    words[0] = lemmatizer.lemmatize(words[0],pos='v')
                words = [word_vocab.get(w,self.UNK) for w in words]+[self.EOS]
            else:
                words = [int(w) for w in words]+[self.EOS]
            padding = [self.PAD for _ in range(query_len - len(words)+2)]
            words.extend(padding)
            self.query.append(words)


    def __len__(self):
        return len(self.query)

    def __getitem__(self, item):
        return torch.tensor(self.query[item]),torch.tensor(self.query[item])


def _compute_loss(model, data_loader, vocab_size, device, with_cuda):
    loss_values = []
    decoded = []

    loss_weight = torch.ones((vocab_size)).to(device)
    loss_weight[0] = 0
    loss_weight[1] = 0

    loss = nn.CrossEntropyLoss(weight=loss_weight)
    for data, target in tqdm.tqdm(data_loader):
        if with_cuda:
            torch.cuda.empty_cache()
        with torch.no_grad():
            data = data.to(device)
            target = target.to(device)
            m, l, z, decoded = model.forward(data)
            decoded = decoded.view(-1, vocab_size)
            loss_value = loss(decoded.view(-1, vocab_size), target.view(-1))
            loss_values.append(loss_value.item())
    return np.array(loss_values)


def model_filter(comments, word_vocab, model, with_cuda=True, query_len=20,num_workers=1,max_iter=1000,dividing_point=None):
    if not isinstance(comments, list):
        raise TypeError('comments must be a list')
    
    if len(comments) < 2:
        raise ValueError('The length of comments must be greater than 1')

    vocab_size = len(word_vocab)
    corpus = QueryDataset(comments, word_vocab, query_len, is_docstring=True)
    data_loader = torch.utils.data.DataLoader(
        dataset=corpus, batch_size=1, shuffle=False, num_workers=num_workers)

    model.eval()
    device = torch.device("cuda:0" if with_cuda else "cpu")
    model = model.to(device)

    losses = _compute_loss(model,data_loader, vocab_size, device, with_cuda)
    idx = np.argsort(losses)
    losses = np.array(losses)[idx].reshape(-1, 1)

    if dividing_point is None:
        gmm = GaussianMixture(
            n_components=2, covariance_type='full', max_iter=max_iter).fit(losses)
        predicted = gmm.predict(losses)
        return_type = predicted[np.argmin(losses)]
        idx_new = idx[predicted == return_type]
        idx_new.sort()
    elif isinstance(dividing_point, int):
        idx_new = idx[:int(len(idx)*dividing_point)]
    elif isinstance(dividing_point, function):
        idx_new = dividing_point(idx,losses)

    filtered_comments = [comments[i] for i in idx_new]
    return filtered_comments, idx_new
